# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.conf import settings
from django.core import checks


def tenant_groups_backend(app_configs, **kwargs):  # pylint: disable=unused-argument
    errors = []
    if "tenant_groups.backends.GroupsBackend" not in settings.AUTHENTICATION_BACKENDS:
        errors.append(
            checks.Error(
                msg="tenant_groups.backends.GroupsBackend "
                "is not pesent in AUTHENTICATION_BACKENDS!",
                hint=(
                    "See "
                    "https://kiwitcms.readthedocs.io/en/latest/admin.html#tenant-groups"
                ),
            )
        )
    return errors



# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django_tenants.utils import (
    get_public_schema_name,
    get_tenant_model,
    tenant_context,
)
from tcms.core.management.commands import refresh_permissions

from tenant_groups.models import Group as TenantGroup


class Command(refresh_permissions.Command):
    help = "Refresh permissions for tenant groups 'Admin' & 'Tester'."

    group_model = TenantGroup
    admin_permissions_filter = {
        "content_type__app_label__in": TenantGroup.relevant_apps,
    }
    tester_permissions_filter = {
        "content_type__app_label__in": TenantGroup.relevant_apps,
    }

    def handle(self, *args, **kwargs):
        for tenant in get_tenant_model().objects.exclude(
            schema_name=get_public_schema_name(),
        ):
            output = None
            if kwargs["verbosity"]:
                output = self.stdout

            with tenant_context(tenant):
                if output:
                    output.write(
                        f"\n\n === Refreshing permissions for tenant '{tenant.schema_name}' ==="
                    )

                self.execute_commands(*args, **kwargs)

                if output:
                    output.write(
                        f"\n\n === End refreshing permissions for tenant '{tenant.schema_name}' ==="
                    )



# Copyright (c) 2019-2020 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import os

from django.conf import settings
from django.utils.functional import cached_property
from django.core.files.storage import FileSystemStorage

from django_tenants import utils


class TenantFileSystemStorage(FileSystemStorage):
    """
    Implementation that extends core Django's FileSystemStorage for multi-tenant setups,
    storing files under induvidual directories. Workaround until
    https://github.com/tomturner/django-tenants/pull/252 gets merged.
    """

    @cached_property
    def relative_media_root(self):  # pylint: disable=no-self-use
        return getattr(settings, "MULTITENANT_RELATIVE_MEDIA_ROOT", "%s")

    @property  # not cached like in parent class
    def base_url(self):  # pylint: disable=invalid-overridden-method
        _url = super().base_url
        _url = os.path.join(
            _url, utils.parse_tenant_config_path(self.relative_media_root)
        )
        if not _url.endswith("/"):
            _url += "/"
        return _url

    @property  # not cached like in parent class
    def location(self):  # pylint: disable=invalid-overridden-method
        _location = os.path.join(
            super().location, utils.parse_tenant_config_path(self.relative_media_root)
        )
        return os.path.abspath(_location)



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import os

from django.db import models
from django.conf import settings

from django_tenants.models import TenantMixin, DomainMixin


class Tenant(TenantMixin):
    auto_create_schema = bool(os.environ.get("AUTO_CREATE_SCHEMA", True))
    auto_drop_schema = True

    name = models.CharField(max_length=100, db_index=True)
    created_on = models.DateTimeField(auto_now_add=True, db_index=True)
    paid_until = models.DateTimeField(null=True, blank=True, db_index=True)
    publicly_readable = models.BooleanField(default=False, db_index=True)
    authorized_users = models.ManyToManyField(to=settings.AUTH_USER_MODEL)
    owner = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE, related_name="tenant_owner"
    )
    organization = models.CharField(max_length=64, null=True, blank=True, db_index=True)
    extra_emails = models.CharField(
        null=True, blank=True, db_index=True, max_length=256
    )

    def __str__(self):
        return f"[{self.schema_name}] {self.name}"


def _authorized_user_str(self):
    return f"{self.user.username}@{self.tenant.name}"


Tenant.authorized_users.through.__str__ = _authorized_user_str


class Domain(DomainMixin):
    def __str__(self):
        return f"{self.domain} for {self.tenant}"



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.conf import settings
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.decorators import permission_required
from django.http import HttpResponseRedirect
from django.views.generic.base import RedirectView
from django.views.generic.edit import FormView, UpdateView
from django.urls import reverse
from django.utils.decorators import method_decorator
from django.utils.translation import gettext_lazy as _

from tcms_tenants import utils
from tcms_tenants.forms import (
    InviteUsersForm,
    NewTenantForm,
    UpdateTenantForm,
    VALIDATION_RE,
)
from tcms_tenants.models import Tenant


@method_decorator(permission_required("tcms_tenants.add_tenant"), name="dispatch")
class NewTenantView(FormView):
    """
    Everybody is allowed to create new tenants.
    Depending on additional configuration they may have to pay
    for using them!
    """

    form_class = NewTenantForm

    def get_template_names(self):
        """
        Allow downstream installations to override the template. For example
        to provide a link to SLA or hosting policy! The overriden template
        can extend the base one if needed.
        """
        return ["tcms_tenants/override_new.html", "tcms_tenants/new.html"]

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tcms_tenants_domain"] = settings.KIWI_TENANTS_DOMAIN
        context["validation_pattern"] = VALIDATION_RE.pattern
        return context

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["initial"]["owner"] = self.request.user.pk
        return kwargs

    def form_valid(self, form):
        tenant = utils.create_tenant(form, self.request)
        # all is successfull so redirect to the new tenant
        return HttpResponseRedirect(utils.tenant_url(self.request, tenant.schema_name))


@method_decorator(permission_required("tcms_tenants.change_tenant"), name="dispatch")
class UpdateTenantView(UpdateView):
    model = Tenant
    form_class = UpdateTenantForm
    success_url = "/"

    def get_template_names(self):
        return ["tcms_tenants/override_edit.html", "tcms_tenants/new.html"]

    @staticmethod
    def check_owner(request):
        # super-user can edit all tenants
        if request.user.is_superuser:
            return None

        # tenant owner can edit their own tenant
        if request.user == request.tenant.owner:
            return None

        messages.add_message(
            request,
            messages.ERROR,
            _("Only super-user and tenant owner are allowed to edit tenant properties"),
        )
        return HttpResponseRedirect("/")

    def get(self, request, *args, **kwargs):
        return self.check_owner(request) or super().get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        return self.check_owner(request) or super().post(request, *args, **kwargs)

    def get_object(self, queryset=None):
        return self.request.tenant

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["tcms_tenants_domain"] = settings.KIWI_TENANTS_DOMAIN
        context["validation_pattern"] = VALIDATION_RE.pattern
        context["form_action_url"] = reverse("tcms_tenants:edit-tenant")
        context["page_title"] = _("Edit tenant")
        return context


@method_decorator(login_required, name="dispatch")
class RedirectTo(RedirectView):  # pylint: disable=missing-permission-required
    """
    Will redirect to tenant.domain.domain/path!

    Used together with GitHub logins and the ``tenant_url``
    template tag!

    When trying to do GitHub login on a tenant sub-domain
    the HTML href will actually point to

    https://public.tenants.localdomain/login/github?next=/tenants/go/to/tenant/path

    instead of

    http://tenant.tenants.localdomain/login/github?next=path

    This is to prevent redirect_uri mismatch errors!
    """

    def get_redirect_url(self, *args, **kwargs):
        tenant = kwargs["tenant"]
        path = kwargs["path"]
        tenant_url = utils.tenant_url(self.request, tenant)
        return f"{tenant_url}/{path}"


@method_decorator(login_required, name="dispatch")
class InviteUsers(FormView):  # pylint: disable=missing-permission-required
    """
    Invite users to tenant via email.

    .. important::

        Anyone who is authorized for this tenant can invite others
        in the same way they can add them directly via the `Authorized users`
        menu!
    """

    form_class = InviteUsersForm
    template_name = "tcms_tenants/invite_users.html"

    def get(self, request, *args, **kwargs):
        if not utils.owns_tenant(request.user, request.tenant):
            messages.add_message(
                request,
                messages.ERROR,
                _("Only users who are authorized for this tenant can invite others"),
            )
            return HttpResponseRedirect("/")

        return super().get(self, request, *args, **kwargs)

    def form_valid(self, form):
        utils.invite_users(self.request, form.cleaned_data["emails"])

        return HttpResponseRedirect("/")



# Copyright (c) 2020-2024 Alexander Todorov <atodorov@otb.bg>
# Copyright (c) 2022 Ivajlo Karabojkov <karabojkov@kitbg.com>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

# pylint: disable=undefined-variable
import os


# start multi-tenant settings override
DATABASES["default"]["ENGINE"] = "django_tenants.postgresql_backend"  # noqa: F821
DATABASE_ROUTERS = [
    "django_tenants.routers.TenantSyncRouter",
]


# attachments storage
DEFAULT_FILE_STORAGE = "tcms_tenants.storage.TenantFileSystemStorage"
MULTITENANT_RELATIVE_MEDIA_ROOT = "tenant/%s"


# this always needs to be the first app
if "django_tenants" not in INSTALLED_APPS:  # noqa: F821
    INSTALLED_APPS.insert(0, "django_tenants")  # noqa: F821

for app_list in (INSTALLED_APPS, TENANT_APPS):  # noqa: F821
    if "tenant_groups" not in app_list:
        app_list.append("tenant_groups")


KIWI_TENANTS_DOMAIN = os.environ.get("KIWI_TENANTS_DOMAIN")


# TenantMainMiddleware always needs to be first
if (
    "django_tenants.middleware.main.TenantMainMiddleware"
    not in MIDDLEWARE  # noqa: F821
):
    MIDDLEWARE.insert(  # noqa: F821
        0, "django_tenants.middleware.main.TenantMainMiddleware"
    )

if (
    "tcms_tenants.middleware.BlockUnauthorizedUserMiddleware"
    not in MIDDLEWARE  # noqa: F821
):
    MIDDLEWARE.append(  # noqa: F821
        "tcms_tenants.middleware.BlockUnauthorizedUserMiddleware"
    )

# replace ModelBackend with GroupsBackend
if "django.contrib.auth.backends.ModelBackend" in AUTHENTICATION_BACKENDS:  # noqa: F821
    idx = AUTHENTICATION_BACKENDS.index(  # noqa: F821
        "django.contrib.auth.backends.ModelBackend"
    )
    AUTHENTICATION_BACKENDS[idx] = "tenant_groups.backends.GroupsBackend"  # noqa: F821

if (
    "tcms_tenants.backends.PubliclyReadableBackend"
    not in AUTHENTICATION_BACKENDS  # noqa: F821
):
    AUTHENTICATION_BACKENDS.append(  # noqa: F821
        "tcms_tenants.backends.PubliclyReadableBackend"
    )


TENANT_MODEL = "tcms_tenants.Tenant"
TENANT_DOMAIN_MODEL = "tcms_tenants.Domain"


# share login session between tenants
SESSION_COOKIE_DOMAIN = f".{KIWI_TENANTS_DOMAIN}"


# everybody can access the main instance
SHARED_APPS = INSTALLED_APPS  # noqa: F821

# extra context processors
for processor in (
    "tcms_tenants.context_processors.schema_name_processor",
    "tcms_tenants.context_processors.tenant_navbar_processor",
):
    if processor not in TEMPLATES[0]["OPTIONS"]["context_processors"]:  # noqa: F821
        TEMPLATES[0]["OPTIONS"]["context_processors"].append(  # noqa: F821
            processor
        )  # noqa: F821



# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html


from django import forms
from django.contrib import admin
from django.contrib import auth
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from tcms_tenants.utils import get_current_tenant
from tenant_groups.models import Group


class GroupAdminForm(forms.ModelForm):
    class Meta:
        model = Group
        fields = ["name", "permissions"]

    users = forms.ModelMultipleChoiceField(
        queryset=auth.get_user_model().objects.none(),
        required=False,
        widget=admin.widgets.FilteredSelectMultiple("users", False),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields["users"].label = capfirst(_("users"))
        self.fields["users"].queryset = get_current_tenant().authorized_users.all()
        if self.instance.pk:
            self.fields["users"].initial = self.instance.user_set.all()

        if "permissions" in self.fields:
            self.fields["permissions"].queryset = auth.models.Permission.objects.filter(
                content_type__app_label__in=Group.relevant_apps
            )
            if self.instance.pk:
                self.fields["permissions"].initial = self.instance.permissions.all()

    def save(self, commit=True):
        instance = super().save(commit=commit)
        instance.save()

        self.instance.user_set.set(self.cleaned_data["users"])
        self.save_m2m()

        return instance


class GroupAdmin(auth.admin.GroupAdmin):
    form = GroupAdminForm

    def has_delete_permission(self, request, obj=None):
        if obj and obj.name in ["Tester", "Administrator"]:
            return False
        return super().has_delete_permission(request, obj)

    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj=obj)
        name_index = fields.index("name")

        # make sure Name is always the first field
        if name_index > 0:
            del fields[name_index]
            fields.insert(0, "name")

        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj)

        if obj and obj.name in ["Tester", "Administrator"]:
            readonly_fields = readonly_fields + ("name",)

        return readonly_fields


admin.site.register(Group, GroupAdmin)



# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

# pylint: disable=import-outside-toplevel

from django.apps import AppConfig as DjangoAppConfig
from django.core.checks import register


class AppConfig(DjangoAppConfig):
    name = "tenant_groups"

    def ready(self):
        from tenant_groups import checks

        register(checks.tenant_groups_backend)



# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.contrib.auth.backends import ModelBackend
from django.contrib.auth.models import Permission

from django_tenants.utils import get_public_schema_name, tenant_context
from tcms_tenants.utils import get_current_tenant


class GroupsBackend(ModelBackend):
    """
    Backend which returns group permissions configured per-tenant but keeps
    individual user permissions!

    NOTE: Individually assigned permissions have higher priority
    (are valid across tenants) compared to group permissions.
    """

    public_schema_name = get_public_schema_name()

    @property
    def tenant(self):
        return get_current_tenant()

    def get_group_permissions(self, user_obj, obj=None):
        # group permissions configured on public tenant via
        # django.contrib.auth.models.Group
        if self.tenant.schema_name == self.public_schema_name:
            return super().get_group_permissions(user_obj, obj)

        # permissions configured per-tenant via
        # tenant_groups.models.Group
        with tenant_context(self.tenant):
            permissions = (
                Permission.objects.filter(tenant_groups__user_set__in=[user_obj])
                .values_list("content_type__app_label", "codename")
                .order_by()
            )

            return {f"{ct}.{name}" for ct, name in permissions}



# Copyright (c) 2019 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tcms_tenants", "0002_tenant_owner"),
    ]

    operations = [
        migrations.AlterField(
            model_name="tenant",
            name="created_on",
            field=models.DateTimeField(auto_now_add=True, db_index=True),
        ),
        migrations.AlterField(
            model_name="tenant",
            name="paid_until",
            field=models.DateTimeField(blank=True, db_index=True, null=True),
        ),
    ]



# Copyright (c) 2021 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tcms_tenants", "0004_tenant_organization"),
    ]

    operations = [
        migrations.RenameField(
            model_name="tenant",
            old_name="on_trial",
            new_name="publicly_readable",
        ),
        # change the defaults
        migrations.AlterField(
            model_name="tenant",
            name="publicly_readable",
            field=models.BooleanField(default=False, db_index=True),
        ),
    ]



# Copyright (c) 2019-2020 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.conf import settings
from django.db import models, migrations
import django_tenants.postgresql_backend.base


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Tenant",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "schema_name",
                    models.CharField(
                        unique=True,
                        max_length=63,
                        db_index=True,
                        validators=[
                            django_tenants.postgresql_backend.base._check_schema_name
                        ],
                    ),
                ),
                ("name", models.CharField(max_length=100, db_index=True)),
                ("created_on", models.DateField(auto_now_add=True, db_index=True)),
                ("paid_until", models.DateField(null=True, blank=True, db_index=True)),
                ("on_trial", models.BooleanField(default=True, db_index=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="tenant",
            name="authorized_users",
            field=models.ManyToManyField(to=settings.AUTH_USER_MODEL),
        ),
        migrations.CreateModel(
            name="Domain",
            fields=[
                (
                    "id",
                    models.AutoField(
                        verbose_name="ID",
                        serialize=False,
                        auto_created=True,
                        primary_key=True,
                    ),
                ),
                (
                    "domain",
                    models.CharField(unique=True, max_length=253, db_index=True),
                ),
                ("is_primary", models.BooleanField(default=True, db_index=True)),
                (
                    "tenant",
                    models.ForeignKey(
                        related_name="domains",
                        to="tcms_tenants.Tenant",
                        on_delete=models.deletion.CASCADE,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



# Copyright (c) 2019 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django import template
from django.db import connection

from tcms_tenants import utils

register = template.Library()


@register.simple_tag
def tenant_url(request, schema_name=None):
    if not schema_name:
        schema_name = connection.schema_name

    return utils.tenant_url(request, schema_name)



# Copyright (c) 2021 Ivajlo Karabojkov <karabojkov@kitbg.com>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

# pylint: disable=import-outside-toplevel
from django.apps import AppConfig as DjangoAppConfig
from django.core.checks import register


class AppConfig(DjangoAppConfig):
    name = "tcms_tenants"

    def ready(self):
        from tcms_tenants import checks

        register(checks.tenants_env_check)



# Copyright (c) 2019-2021 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.urls import re_path
from tcms_tenants import views

app_name = "tcms_tenants"

urlpatterns = [
    re_path(r"^create/$", views.NewTenantView.as_view(), name="create-tenant"),
    re_path(r"^edit/$", views.UpdateTenantView.as_view(), name="edit-tenant"),
    re_path(r"^invite/$", views.InviteUsers.as_view(), name="invite-users"),
    re_path(
        r"^go/to/(?P<tenant>\w+)/(?P<path>.*)$",
        views.RedirectTo.as_view(),
        name="redirect-to",
    ),
]



# Copyright (c) 2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("tcms_tenants", "0005_rename_on_trial_to_public_read"),
    ]

    operations = [
        migrations.AddField(
            model_name="tenant",
            name="extra_emails",
            field=models.CharField(
                blank=True, db_index=True, max_length=256, null=True
            ),
        ),
    ]



# Copyright (c) 2021-2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.contrib.auth.backends import BaseBackend

from tcms_tenants.utils import get_current_tenant


class PubliclyReadableBackend(BaseBackend):
    """
    Backend which always returns True for *.view_* permissions for
    publicly readable tenants!
    """

    def has_perm(self, user_obj, perm, obj=None):
        # let other backends decide
        if perm.startswith("auth."):
            return False

        current_tenant = get_current_tenant()

        # allow everyone to view
        if current_tenant.publicly_readable and perm.find(".view_") > -1:
            return True

        return super().has_perm(user_obj, perm, obj=obj)



# Copyright (c) 2019-2022 Alexander Todorov <atodorov@MrSenko.com>

# Licensed under the GPL 3.0: https://www.gnu.org/licenses/gpl-3.0.txt

from django.db import connection, transaction
from django.db.utils import ProgrammingError
from django.conf import settings

import factory
from factory.django import DjangoModelFactory

from django_tenants.clone import CloneSchema
from django_tenants.test.client import TenantClient
from django_tenants.test.cases import FastTenantTestCase
from django_tenants.utils import get_tenant_model, schema_context, tenant_context

from tcms.tests.factories import TestPlanFactory
from tenant_groups.models import Group as TenantGroup


class UserFactory(DjangoModelFactory):
    class Meta:
        model = settings.AUTH_USER_MODEL

    # avoid conflicts with tcms.tests.factories.UserFactory
    username = factory.Sequence(lambda n: f"tenant-user{n}")
    first_name = factory.Sequence(lambda n: f"First{n}")
    last_name = factory.Sequence(lambda n: f"Last{n}-ov")
    email = factory.LazyAttribute(lambda user: f"{user.username}@kiwitcms.org")
    is_active = True
    is_staff = True
    is_superuser = False


class LoggedInTestCase(FastTenantTestCase):
    @classmethod
    def get_test_schema_name(cls):
        return "fast"

    @classmethod
    def setup_tenant(cls, tenant):
        super().setup_tenant(tenant)

        tenant.publicly_readable = False
        tenant.owner = UserFactory()
        tenant.owner.set_password("password")
        tenant.owner.save()

    @classmethod
    def setUpClass(cls):
        # b/c it may create a new tenant
        with schema_context("public"):
            super().setUpClass()

        # authorize tenant owner
        cls.tenant.authorized_users.add(cls.tenant.owner)

        cls.tester = UserFactory()
        cls.tester.set_password("password")
        cls.tester.save()

        # authorize this user
        cls.tenant.authorized_users.add(cls.tester)

        # initial data
        with tenant_context(cls.tenant):
            cls.test_plan_by_owner = TestPlanFactory(author=cls.tenant.owner)

    def setUp(self):
        super().setUp()

        self.client = TenantClient(self.tenant)
        self.client.login(
            username=self.tester.username,  # nosec:B106:hardcoded_password_funcarg
            password="password",
        )


class TenantGroupsTestCase(LoggedInTestCase):
    original_value = None

    @classmethod
    def get_test_schema_name(cls):
        # NB: this is a special name, utils.create_tenant() will
        # use cloning if this schema exists which is faster
        return "empty"

    @classmethod
    def get_test_tenant_domain(cls):
        return "empty.test.com"

    @classmethod
    def get_verbosity(cls):
        return 1

    @classmethod
    def setUpClass(cls):
        # will create the actual schema
        cls.original_value = get_tenant_model().auto_create_schema
        get_tenant_model().auto_create_schema = True

        # execute before calling the parent class which will start a DB transaction
        with schema_context("public"):
            cursor = connection.cursor()

            try:
                cursor.execute("SELECT 'clone_schema'::regproc")
            except ProgrammingError:
                CloneSchema()._create_clone_schema_function()  # pylint: disable=protected-access
                transaction.commit()

            # tenant creation may happen here so it needs to be on public schema
            super().setUpClass()

        with tenant_context(cls.tenant):
            # add owner to default groups b/c they need certain permissions
            # and b/c this is what utils.create_tenant() does
            TenantGroup.objects.get(name="Administrator").user_set.add(cls.tenant.owner)
            TenantGroup.objects.get(name="Tester").user_set.add(cls.tenant.owner)

    @classmethod
    def tearDownClass(cls):
        get_tenant_model().auto_create_schema = cls.original_value
        super().tearDownClass()



# Copyright (c) 2020-2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _


# Follows the format of ``tcms.settings.common.MENU_ITEMS``
MENU_ITEMS = [
    (
        _("Tenant"),
        [
            (_("Create"), reverse_lazy("tcms_tenants:create-tenant")),
            (_("Edit"), reverse_lazy("tcms_tenants:edit-tenant")),
            ("-", "-"),
            (_("Groups"), "/admin/tenant_groups/group/"),
            ("-", "-"),
            (_("Invite users"), reverse_lazy("tcms_tenants:invite-users")),
            (_("Authorized users"), "/admin/tcms_tenants/tenant_authorized_users/"),
        ],
    ),
]



# Copyright (c) 2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import datetime

from django.contrib.auth import get_user_model
from django.core.management import call_command
from django.core.management.base import BaseCommand
from django.utils import timezone


class Command(BaseCommand):
    help = "Initialize tenants for a *NEW* Kiwi TCMS Enterprise installation."

    def add_arguments(self, parser):
        parser.add_argument(
            "--domain",
            action="store",
            dest="domain",
            required=True,
            help="The primary tenant domain",
        )

    def handle(self, *args, **kwargs):
        superuser = get_user_model().objects.filter(is_superuser=True).first()
        paid_until = timezone.now() + datetime.timedelta(days=100 * 365)
        domain = kwargs["domain"]

        call_command(
            "create_tenant",
            f"--verbosity={kwargs['verbosity']}",
            "--schema_name",
            "public",
            "--name",
            "Public tenant",
            "--paid_until",
            paid_until.isoformat(),
            "--publicly_readable",
            False,
            "--owner_id",
            superuser.pk,
            "--organization",
            "Testing department",
            "--extra_emails",
            superuser.email,
            "--domain-domain",
            domain,
            "--domain-is_primary",
            True,
        )

        # a special tenant to make cloning faster
        call_command(
            "create_tenant",
            f"--verbosity={kwargs['verbosity']}",
            "--schema_name",
            "empty",
            "--name",
            "Cloning Template",
            "--paid_until",
            paid_until.isoformat(),
            "--publicly_readable",
            False,
            "--owner_id",
            superuser.pk,
            "--organization",
            "Kiwi TCMS",
            "--extra_emails",
            superuser.email,
            "--domain-domain",
            "empty.example.org",
            "--domain-is_primary",
            True,
        )



# Copyright (c) 2019 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("tcms_tenants", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="tenant",
            name="owner",
            field=models.ForeignKey(
                on_delete=models.CASCADE,
                related_name="tenant_owner",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



# Copyright (c) 2022 Ivajlo Karabojkov <karabojkov@kitbg.com>
# Copyright (c) 2022-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from attachments.models import Attachment
from django.template.loader import render_to_string


def tenant_navbar_processor(request):
    """
    Provide tenant name for display in navbar
    """
    if not getattr(request, "tenant", None):
        return {"CUSTOMIZED_LOGO_CONTENTS": ""}

    if request.tenant.organization:
        tenant_name = request.tenant.organization
    elif request.tenant.name:
        tenant_name = request.tenant.name
    else:
        tenant_name = request.tenant.schema_name

    tenant_logo = Attachment.objects.attachments_for_object(request.tenant).first()

    customized_logo_contents = render_to_string(
        "tcms_tenants/tenant_name.html",
        {"tenant_name": tenant_name, "tenant_logo": tenant_logo},
    )
    return {"CUSTOMIZED_LOGO_CONTENTS": customized_logo_contents}


def schema_name_processor(request):
    """
    Provide tenant.schema_name for use in templates
    """
    return {
        "TENANT_SCHEMA_NAME": getattr(
            getattr(request, "tenant", None), "schema_name", ""
        )
    }



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import re

from django import forms
from django.forms.utils import ErrorList
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _

from django_tenants.postgresql_backend import base as validators

from tcms.core.utils.mailto import custom_email_validators
from tcms_tenants import models
from tcms_tenants import utils


VALIDATION_RE = re.compile(r"^[a-z0-9]{1,63}$")


def _check_domain_name(name):
    """
    Because we use schema_name as sub-domains interchangeably
    we impose additional restriction on input values:
        alpha-numeric, 63 chars max
    """
    if not VALIDATION_RE.match(name):
        raise ValidationError(_("Invalid string"))


class NewTenantForm(forms.ModelForm):
    class Meta:
        model = models.Tenant
        exclude = ("authorized_users",)  # pylint: disable=modelform-uses-exclude

    schema_name = forms.CharField(
        max_length=63,
        # pylint: disable=protected-access
        validators=[
            validators._check_schema_name,
            _check_domain_name,
            utils.schema_name_not_used,
        ],
    )
    paid_until = forms.DateTimeField(required=False, widget=forms.HiddenInput)
    organization = forms.CharField(
        max_length=64, required=False, widget=forms.HiddenInput
    )
    extra_emails = forms.CharField(
        max_length=256, required=False, widget=forms.HiddenInput
    )


class InviteUsersForm(forms.Form):  # pylint: disable=must-inherit-from-model-form
    number_of_fields = 10

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for i in self.range:
            self.fields[f"email_{i}"] = forms.EmailField(
                required=False,
                validators=[custom_email_validators],
            )

    def clean(self):
        emails = set()

        for i in self.range:
            email = self.cleaned_data.get(f"email_{i}")
            if email:
                emails.add(email)

        self.cleaned_data["emails"] = emails

    @property
    def range(self):
        return range(self.number_of_fields)


class UpdateTenantForm(NewTenantForm):
    enabled_fields = (
        "name",
        "publicly_readable",
        "extra_emails",
    )

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        data=None,
        files=None,
        auto_id="id_%s",
        prefix=None,
        initial=None,
        error_class=ErrorList,
        label_suffix=None,
        empty_permitted=False,
        instance=None,
        use_required_attribute=None,
        renderer=None,
    ):
        super().__init__(
            data,
            files,
            auto_id,
            prefix,
            initial,
            error_class,
            label_suffix,
            empty_permitted,
            instance,
            use_required_attribute,
            renderer,
        )

        for field in self.fields:
            self.fields[field].disabled = field not in self.enabled_fields

        # remove validator which only makes sense when creating new tenants
        self.fields["schema_name"].validators.remove(utils.schema_name_not_used)



# Copyright (c) 2022-2023 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import datetime

from django.contrib.auth import get_user_model

from tcms_tenants.forms import NewTenantForm
from tcms_tenants.utils import create_tenant


def create_oss_tenant(owner, name, schema_name, organization):
    """
    Used to create tenants for our OSS program. Executed by the
    instance administrator!
    """

    class FakeRequest:  # pylint: disable=too-few-public-methods,nested-class-found
        is_secure = True
        user = None

        def __init__(self, username):
            self.user = get_user_model().objects.get(username=username)

    request = FakeRequest(owner)

    form = NewTenantForm(
        {
            "owner": request.user.pk,
            "name": name,
            "schema_name": schema_name.lower(),
            "organization": organization,
            "publicly_readable": False,
            "paid_until": datetime.datetime(2999, 12, 31),
        }
    )
    if form.is_valid():
        return create_tenant(form, request)

    raise RuntimeError(list(form.errors.items()))



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

# pylint: disable=too-few-public-methods
from datetime import timedelta

from django.contrib import messages
from django.http import HttpResponseForbidden
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from tcms_tenants.utils import can_access


class BlockUnauthorizedUserMiddleware:
    """
    Raises 403 if the user making the request is not authorized
    explicitly to access the tenant instance!

    .. warning:

        This must be placed after
        ``django_tenants.middleware.main.TenantMainMiddleware`` and
        ``django.contrib.auth.middleware.AuthenticationMiddleware`` -
        usually goes at the end
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):  # pylint: disable=no-self-use
        if not can_access(request.user, request.tenant):
            return HttpResponseForbidden(_("Unauthorized"))

        return self.get_response(request)


class BlockUnpaidTenantMiddleware:
    """
    Blocks unpaid tenants or adds warning messages if tenant is about to
    expire soon!

    .. warning:

        This must be placed after
        ``tcms_tenants.middleware.BlockUnauthorizedUserMiddleware`` -
        usually goes at the end.
    """

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        if (request.tenant.paid_until is None) or (
            request.tenant.paid_until <= timezone.now()
        ):
            return HttpResponseForbidden(_("Unpaid"))

        if request.tenant.paid_until <= timezone.now() + timedelta(days=1):
            for msg in messages.get_messages(request):
                if msg.level_tag == "warning":
                    break
            else:
                # will be shown only if no other warnings are present
                messages.add_message(
                    request,
                    messages.WARNING,
                    _("Tenant expires soon"),
                    fail_silently=True,
                )

        return self.get_response(request)



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

# pylint: disable=missing-docstring

from setuptools import setup, find_packages


def get_long_description():
    with open("README.rst", "r", encoding="utf-8") as file:
        return file.read()


def get_install_requires(path):
    requires = []

    with open(path, "r", encoding="utf-8") as file:
        for line in file:
            if line.startswith("-r "):
                continue
            requires.append(line.strip())
        return requires


setup(
    name="kiwitcms-tenants",
    version="3.2.2",
    description="Multi-tenant support for Kiwi TCMS",
    long_description=get_long_description(),
    author="Kiwi TCMS",
    author_email="info@kiwitcms.org",
    url="https://github.com/kiwitcms/tenants/",
    license="AGPLv3+",
    install_requires=get_install_requires("requirements.txt"),
    include_package_data=True,
    packages=find_packages(exclude=["test_project*", "*.tests"]),
    zip_safe=False,
    entry_points={"kiwitcms.plugins": ["kiwitcms_tenants = tcms_tenants"]},
    classifiers=[
        "Framework :: Django",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.11",
    ],
)



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django import forms
from django.urls import reverse
from django.contrib import admin
from django.forms.utils import ErrorList
from django.utils.translation import gettext_lazy as _
from django.http import HttpResponseForbidden, HttpResponseRedirect

from tcms.core.forms.fields import UserField

from tcms_tenants.models import Tenant
from tcms_tenants.utils import add_to_default_groups, owns_tenant, tenant_url


class TenantAdmin(admin.ModelAdmin):
    """
    Allows only super-user to see and delete tenants!
    """

    actions = ["delete_selected"]
    list_display = (
        "id",
        "name",
        "schema_name",
        "domain_name",
        "created_on",
        "publicly_readable",
        "paid_until",
        "owner",
        "extra_emails",
        "organization",
    )
    search_fields = ("name", "schema_name", "organization")
    ordering = ["-created_on"]

    def add_view(self, request, form_url="", extra_context=None):
        return HttpResponseRedirect(reverse("tcms_tenants:create-tenant"))

    def change_view(self, request, object_id, form_url="", extra_context=None):
        tenant = Tenant.objects.get(pk=object_id)
        target_url = tenant_url(request, tenant.schema_name) + reverse(
            "tcms_tenants:edit-tenant"
        )
        return HttpResponseRedirect(target_url)

    @admin.options.csrf_protect_m
    def changelist_view(self, request, extra_context=None):
        if request.user.is_superuser:
            return super().changelist_view(request, extra_context)

        return HttpResponseForbidden(_("Unauthorized"))

    @admin.options.csrf_protect_m
    def delete_view(self, request, object_id, extra_context=None):
        if request.user.is_superuser:
            return super().delete_view(request, object_id, extra_context)

        return HttpResponseForbidden(_("Unauthorized"))

    def domain_name(self, instance):  # pylint: disable=no-self-use
        return instance.domains.filter(is_primary=True).first().domain


class AuthorizedUsersChangeForm(forms.ModelForm):
    """
    A custom add/change form which will filter the available
    values of the ``tenant`` field so that only the current
    tenant is shown!
    """

    # IMPORTANT NOTICE:
    # As a security concern we would like to have this queryset set to
    # Tenant.objects.none() instead of .all(). However ModelChoiceField.to_python()
    # is trying to self.queryset.get() the currently selected value! When the queryset
    # is empty this raises ValidationError!
    # The internal mechanics of this are in BaseForm._clean_fields()::L399(Django 2.1.7)
    # which calls field.clean(value) before any clean_<field_name>() methods on the form!
    tenant = forms.models.ModelChoiceField(
        queryset=Tenant.objects.all(),
    )
    user = UserField(  # pylint: disable=form-field-help-text-used
        help_text=_("Existing username, email or user ID")
    )

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
        self,
        data=None,
        files=None,
        auto_id="id_%s",
        prefix=None,
        initial=None,
        error_class=ErrorList,
        label_suffix=None,
        empty_permitted=False,
        instance=None,
        use_required_attribute=None,
        renderer=None,
    ):
        super().__init__(
            data,
            files,
            auto_id,
            prefix,
            initial,
            error_class,
            label_suffix,
            empty_permitted,
            instance,
            use_required_attribute,
            renderer,
        )
        # this is passed by ModelAdmin._chageform_view():L1578
        # when adding a new object
        if initial and "tenant" in initial:
            self.fields["tenant"].queryset = Tenant.objects.filter(pk=initial["tenant"])
        # this is passed by ModelAdmin._chageform_view():L1581
        # when changing an existing object
        elif instance:
            self.fields["tenant"].queryset = Tenant.objects.filter(
                pk=instance.tenant.pk
            )

    def save(self, commit=True):
        instance = super().save(commit=commit)
        instance.save()

        add_to_default_groups(self.instance.user)
        return instance


class AuthorizedUsersAdmin(admin.ModelAdmin):
    """
    Allows administering which users are authorized for tenants!
    """

    actions = ["delete_selected"]
    list_display = (
        "user_username",
        "user_full_name",
        "groups",
    )
    search_fields = ("user__username",)

    form = AuthorizedUsersChangeForm

    def user_username(self, instance):  # pylint: disable=no-self-use
        return instance.user.username

    user_username.short_description = _("Username")
    user_username.admin_order_field = "user__username"

    def user_full_name(self, instance):  # pylint: disable=no-self-use
        return instance.user.get_full_name()

    user_full_name.short_description = _("Full name")

    def groups(self, instance):  # pylint: disable=no-self-use
        return list(instance.user.tenant_groups.values_list("name", flat=True))

    groups.short_description = _("Groups")

    def get_queryset(self, request):
        """
        Show only users authorized for the current tenant!
        """
        return super().get_queryset(request).filter(tenant=request.tenant)

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def render_change_form(
        self, request, context, add=False, change=False, form_url="", obj=None
    ):
        context["adminform"].form.fields["tenant"].queryset = Tenant.objects.filter(
            pk=request.tenant.pk
        )

        return super().render_change_form(
            request, context, add=add, change=change, form_url=form_url, obj=obj
        )

    def get_changeform_initial_data(self, request):
        """
        Pass the current tenant to AuthorizedUsersChangeForm.__init__()
        which will filter the available tenants and only show the current
        one when add/change objects!
        """
        return {"tenant": request.tenant.pk}

    def has_add_permission(self, request):
        """
        Allow to add new authorized users.
        """
        return owns_tenant(request.user, request.tenant)

    def has_change_permission(self, request, obj=None):
        """
        Allow to display the list of of authorized users.
        """
        return owns_tenant(request.user, request.tenant)

    def has_delete_permission(self, request, obj=None):
        """
        Allow to delete selected users.
        """
        return request.user.is_superuser or owns_tenant(request.user, request.tenant)

    def has_module_permission(self, request):
        """
        Allow this module to be seen in main admin page.
        """
        return owns_tenant(request.user, request.tenant)

    def get_model_perms(self, request):
        """
        Allow this module to be seen in main admin page.
        """
        if request.user.is_superuser:
            return super().get_model_perms(request)

        return {"view": owns_tenant(request.user, request.tenant)}

    def delete_model(self, request, obj):
        """
        Remove user from tenant groups before removing the authorization!
        """
        for group in obj.user.tenant_groups.all():
            group.user_set.remove(obj.user)

        super().delete_model(request, obj)


admin.site.register(Tenant, TenantAdmin)
admin.site.register(Tenant.authorized_users.through, AuthorizedUsersAdmin)



# Copyright (c) 2022 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.conf import settings
from django.db import connections
from django.db import migrations, models

from django_tenants.utils import (
    get_public_schema_name,
    get_tenant_database_alias,
)


def forwards_add_default_groups_and_permissions(apps, schema_editor):
    current_tenant = connections[get_tenant_database_alias()].tenant
    if current_tenant.schema_name == get_public_schema_name():
        return

    group_model = apps.get_model("tenant_groups", "Group")
    group_model.objects.bulk_create(
        [
            group_model(name="Administrator"),
            group_model(name="Tester"),
        ]
    )

    permission_model = apps.get_model("auth", "Permission")

    tester = group_model.objects.get(name="Tester")
    tester_perms = tester.permissions.all()
    # apply all permissions for test case & product management
    for app_name in [
        "bugs",
        "django_comments",
        "linkreference",
        "management",
        "testcases",
        "testplans",
        "testruns",
        "attachments",
    ]:
        app_perms = permission_model.objects.filter(
            content_type__app_label__contains=app_name
        )
        app_perms = app_perms.exclude(pk__in=tester_perms).exclude(
            content_type__app_label="attachments",
            codename="delete_foreign_attachments",
        )
        tester.permissions.add(*app_perms)

    # Admin gets the same permissions as Tester + tenant_groups
    admin = group_model.objects.get(name="Administrator")
    admin.permissions.add(*tester.permissions.all())
    admin.permissions.add(
        *permission_model.objects.filter(content_type__app_label="tenant_groups")
    )


def forwards_add_authorized_users_to_default_groups(apps, schema_editor):
    current_tenant = connections[get_tenant_database_alias()].tenant
    if current_tenant.schema_name in [get_public_schema_name(), "empty"]:
        return

    group_model = apps.get_model("tenant_groups", "Group")
    tenant_model = apps.get_model("tcms_tenants", "Tenant")

    tester = group_model.objects.get(name="Tester")
    administrator = group_model.objects.get(name="Administrator")

    # Replace FakeTenant with Tenant b/c we need to access .owner
    current_tenant = tenant_model.objects.get(schema_name=current_tenant.schema_name)

    # make sure that existing users are assigned to default groups
    # to preserve older behavior
    administrator.user_set.add(current_tenant.owner)
    tester.user_set.add(*current_tenant.authorized_users.all())


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("auth", "0012_alter_user_first_name_max_length"),
        ("tcms_tenants", "0005_rename_on_trial_to_public_read"),
    ]

    operations = [
        migrations.CreateModel(
            name="Group",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(max_length=150, unique=True, verbose_name="name"),
                ),
                (
                    "permissions",
                    models.ManyToManyField(
                        blank=True,
                        related_name="tenant_groups",
                        to="auth.Permission",
                        verbose_name="permissions",
                    ),
                ),
                (
                    "user_set",
                    models.ManyToManyField(
                        to=settings.AUTH_USER_MODEL, related_name="tenant_groups"
                    ),
                ),
            ],
            options={
                "verbose_name": "group",
                "verbose_name_plural": "groups",
            },
        ),
        migrations.RunPython(
            code=forwards_add_default_groups_and_permissions,
            reverse_code=migrations.RunPython.noop,
        ),
        migrations.RunPython(
            code=forwards_add_authorized_users_to_default_groups,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



# Copyright (c) 2022-2023 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

from django.db import models
from django.conf import settings
from django.contrib.auth.models import Permission
from django.utils.translation import gettext_lazy as _


class Group(models.Model):
    """
    Per-tenant group definition! See
    https://github.com/kiwitcms/tenants/issues/104#issuecomment-949956789

    Note: essentially copy-pasted from django.contrib.auth b/c we need separate
    records inside tenant's schema.
    """

    relevant_apps = [
        "attachments",
        "bugs",
        "django_comments",
        "linkreference",
        "management",
        "testcases",
        "testplans",
        "testruns",
        "tenant_groups",
        "trackers_integration",
    ]

    name = models.CharField(_("name"), max_length=150, unique=True)
    permissions = models.ManyToManyField(
        Permission,
        verbose_name=_("permissions"),
        blank=True,
        related_name="tenant_groups",
    )

    user_set = models.ManyToManyField(
        to=settings.AUTH_USER_MODEL, related_name="tenant_groups"
    )

    class Meta:
        verbose_name = _("group")
        verbose_name_plural = _("groups")

    def __str__(self):
        return self.name

    def natural_key(self):
        return (self.name,)



# Copyright (c) 2019-2024 Alexander Todorov <atodorov@otb.bg>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import datetime
import uuid

from django.conf import settings
from django.db import connections
from django.contrib import messages
from django.contrib.sites.models import Site
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management import call_command
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.contrib.contenttypes.models import ContentType

from django_tenants.utils import (
    get_public_schema_name,
    get_tenant_database_alias,
    schema_context,
    tenant_context,
)
from tcms.kiwi_auth import forms as kiwi_auth_forms
from tcms.core.utils.mailto import mailto

from tcms_tenants.models import Domain, Tenant
from tenant_groups.models import Group as TenantGroup


UserModel = get_user_model()


def get_current_tenant():
    return connections[get_tenant_database_alias()].tenant


def can_access(user, tenant):
    # everybody can access the public schema
    if tenant.schema_name == get_public_schema_name():
        return True

    # everyone can access publicly readable tenants
    if tenant.publicly_readable:
        return True

    # anonymous users "can access" because they need to be able
    # to reach the login page
    if not user.is_authenticated:
        return True

    return tenant.authorized_users.filter(pk=user.pk).exists()


def owns_tenant(user, tenant):
    return (
        tenant.schema_name != get_public_schema_name()
        and tenant.authorized_users.filter(pk=user.pk).exists()
    )


def schema_name_not_used(name):
    if Tenant.objects.filter(schema_name=name).exists():
        raise ValidationError(_("Schema name already in use"))


# warning: doesn't play well when the domain has a port number
def tenant_domain(schema_name):
    # take into account the fact that some customers deploy their 'public' schema
    # without a prefix, e.g. tcms.example.com == KIWI_TENANTS_DOMAIN or could use a
    # different domain for their tenant(s)!
    domain = Domain.objects.filter(
        tenant__schema_name=schema_name, is_primary=True
    ).first()
    if domain:
        return domain.domain

    return f"{schema_name}.{settings.KIWI_TENANTS_DOMAIN}"


def tenant_url(request, schema_name):
    url = tenant_domain(schema_name)
    if request.is_secure:
        url = "https://" + url
    else:
        url = "http://" + url
    return url


def create_tenant(form, request):
    with schema_context("public"):
        # If a schema with name "empty" exists then use it for
        # cloning b/c that's faster
        if Tenant.objects.filter(schema_name="empty").first():
            schema_name = form.cleaned_data["schema_name"]
            paid_until = form.cleaned_data["paid_until"] or datetime.datetime(
                3000, 3, 31
            )
            call_command(
                "clone_tenant",
                "--clone_from",
                "empty",
                "--clone_tenant_fields",
                False,
                "--schema_name",
                schema_name,
                "--name",
                form.cleaned_data["name"],
                "--paid_until",
                paid_until,
                "--publicly_readable",
                form.cleaned_data["publicly_readable"],
                "--owner_id",
                request.user.pk,
                "--organization",
                form.cleaned_data["organization"] or "-",
                "--extra_emails",
                form.cleaned_data["extra_emails"] or "-",
                "--domain-domain",
                tenant_domain(schema_name),
                "--domain-is_primary",
                True,
            )
            tenant = Tenant.objects.get(schema_name=schema_name)
            if not form.cleaned_data["paid_until"]:
                tenant.paid_until = None
                tenant.save()

            if not form.cleaned_data["organization"]:
                tenant.organization = ""
                tenant.save()

            domain = tenant.domains.first()
        else:
            # otherwise default to applying all migrations one by one
            tenant = form.save()
            domain = Domain.objects.create(
                domain=tenant_domain(tenant.schema_name),
                is_primary=True,
                tenant=tenant,
            )

        # work-around: for some reason the ContentType for tenant-user
        # relationship model doesn't get automatically created and if
        # the owner tries to add other authorized users via admin the
        # action will fail b/c ModelAdmin.log_addition tries to add a
        # a logging record linking to this content type and fails !
        ContentType.objects.get_for_model(tenant.authorized_users.through)

        # make the owner the first authorized user
        # otherwise they can't login
        tenant.authorized_users.add(tenant.owner)

    with tenant_context(tenant):
        # this is used to build full URLs for things like emails
        site = Site.objects.get(pk=settings.SITE_ID)
        site.domain = domain.domain
        site.name = domain.domain
        site.save()

        # add owner to default groups b/c they need certain permissions
        TenantGroup.objects.get(name="Administrator").user_set.add(tenant.owner)
        TenantGroup.objects.get(name="Tester").user_set.add(tenant.owner)

    mailto(
        template_name="tcms_tenants/email/new.txt",
        recipients=[tenant.owner.email],
        subject=str(_("New Kiwi TCMS tenant created")),
        context={
            "tenant_url": tenant_url(request, tenant.schema_name),
        },
    )
    return tenant


# NOTE: defined here to avoid circular imports with forms.py
class RegistrationForm(
    kiwi_auth_forms.RegistrationForm
):  # pylint: disable=too-many-ancestors
    """
    Override captcha field b/c Kiwi TCMS may be configured to
    use reCAPTCHA and we don't want this to block automatic
    creation of user accounts!
    """

    captcha = None


def create_user_account(email_address):
    desired_username = username = email_address.split("@")[0]
    password = uuid.uuid4().hex

    i = 1
    while UserModel.objects.filter(username=username).exists():
        username = f"{desired_username}.{i}"
        i += 1

    form = RegistrationForm(
        data={
            "username": username,
            "password1": password,
            "password2": password,
            "email": email_address,
        }
    )

    user = form.save()

    # activate their account instead of sending them email with activation key
    user.is_active = True
    user.save()

    return user


def add_to_default_groups(user, request=None):
    """
    If there are tenant groups whose names match the default setting the user
    will be added to them. Otherwise it's up to administrator to assign
    appropriate membership & permissions for authorized users.
    """
    for group in TenantGroup.objects.filter(name__in=settings.DEFAULT_GROUPS):
        user.tenant_groups.add(group)
        if request:
            messages.add_message(
                request,
                messages.SUCCESS,
                _(f"User {user.username} added to tenant group {group.name}"),
            )


def invite_users(request, email_addresses):
    the_tenant_url = tenant_url(request, request.tenant.schema_name).strip("/")
    pwd_reset_path = reverse_lazy("tcms-password_reset").rstrip("/")
    email_context = {
        "invited_by": request.user.get_full_name() or request.user.username,
        "tenant_url": the_tenant_url,
        "password_reset_url": f"{the_tenant_url}/{pwd_reset_path}",
    }

    for email in email_addresses:
        # note: users are on public_schema
        user = UserModel.objects.filter(email=email).first()

        # email not found, need to create account for them
        if not user:
            user = create_user_account(email)

        # user already authorized for tenant
        if request.tenant.authorized_users.filter(pk=user.pk).exists():
            continue

        request.tenant.authorized_users.add(user)
        with tenant_context(request.tenant):
            add_to_default_groups(user)

        mailto(
            template_name="tcms_tenants/email/invite_user.txt",
            recipients=[user.email],
            subject=str(_("Invitation to join Kiwi TCMS")),
            context=email_context,
        )



# Copyright (c) 2021 Ivajlo Karabojkov <karabojkov@kitbg.com>
#
# Licensed under GNU Affero General Public License v3 or later (AGPLv3+)
# https://www.gnu.org/licenses/agpl-3.0.html

import os

from django.core import checks


def tenants_env_check(app_configs, **kwargs):  # pylint: disable=unused-argument
    errors = []
    if not os.environ.get("KIWI_TENANTS_DOMAIN"):
        errors.append(
            checks.Error(
                msg="KIWI_TENANTS_DOMAIN environment variable is not set!",
                hint=(
                    "This variable is mandatory, "
                    "see https://github.com/kiwitcms/tenants#installation"
                ),
            )
        )
    return errors
